from stealth25519.ed25519 import point_decompress, secret_to_public, secret_expand


class PublicKey:
    """
    Represents a public key for Ed25519-based cryptography.

    Attributes:
        publicBytes (bytes): The raw bytes of the public key.
        B (Point): The point on the curve corresponding to the public key.

    Methods:
        getPublicBytes(): Get the raw bytes of the public key.
        getPointOnCurve(): Get the point on the curve corresponding to the public key.
        __str__(): Return a hexadecimal representation of the public key.
    """

    def __init__(self, publicBytes = None):
        """
        Initialize a PublicKey instance.

        Args:
            publicBytes (bytes): The raw bytes of the public key.
        """

        self.publicBytes = publicBytes
        self.B = point_decompress(publicBytes)

    def getPublicBytes(self):
        """Get the raw bytes of the public key."""

        return self.publicBytes

    def getPointOnCurve(self):
        """Get the point on the curve corresponding to the public key."""

        return self.B

    def __str__(self):
        """Return a hexadecimal representation of the public key."""

        return self.publicBytes.hex()



class PrivateKey:
    """
    Represents a private key for Ed25519-based cryptography.

    Attributes:
        privateBytes (bytes): The raw bytes of the private key.
        b (int): The private scalar used in key generation.
        publicKey (PublicKey): The corresponding public key.

    Methods:
        generatePublicKey(): Generate the corresponding public key.
        getPrivateBytes(): Get the raw bytes of the private key.
        __str__(): Return a hexadecimal representation of the private key.
    """

    def __init__(self, privateBytes = None):
        """
        Initialize a PrivateKey instance.

        Args:
            privateBytes (bytes): The raw bytes of the private key.
        """

        publicBytes = secret_to_public(privateBytes) # ed25519
        b, dummy = secret_expand(privateBytes) # ed25519
        
        self.privateBytes = privateBytes
        self.b = b
        self.publicKey = PublicKey(publicBytes)

    def generatePublicKey(self):
        """Generate the corresponding public key."""

        return self.publicKey

    def getPrivateBytes(self):
        """Get the raw bytes of the private key."""

        return self.privateBytes

    def __str__(self):
        """Return a hexadecimal representation of the private key."""

        return self.privateBytes.hex()



from stealth25519.ed25519 import point_decompress


class StealthAddress:
    """
    Represents a Stealth Address for Ed25519-based cryptography.

    Attributes:
        Rs (bytes): The raw bytes of the compressed R point.
        Ps (bytes): The raw bytes of the compressed P point.
        R (Point): The decompressed R point.

    Methods:
        __str__(): Return a formatted string representation of the Stealth Address.
    """

    def __init__(self, R = None, P = None):
        """
        Initialize a StealthAddress instance.

        Args:
            R (bytes): The raw bytes of the compressed R point.
            P (bytes): The raw bytes of the compressed P point.
        """

        self.Rs = R
        self.Ps = P
        self.R = point_decompress(R) # ed25519
    
    def __str__(self):
        """Return a formatted string representation of the Stealth Address."""

        return f'R: {self.Rs.hex()}\nP: {self.Ps.hex()}'




# RFC8032: https://datatracker.ietf.org/doc/html/rfc8032

import hashlib

p = 2**255 - 19
q = 2**252 + 27742317777372353535851937790883648493


def modp_inv(x):
    return pow(x, p-2, p)


d = -121665 * modp_inv(121666) % p
modp_sqrt_m1 = pow(2, (p-1) // 4, p)


def sha512(s):
    return hashlib.sha512(s).digest()


def sha256(s):
    return hashlib.sha256(s).digest()


def sha512_modq(s):
    return int.from_bytes(sha512(s), "little") % q

def bytes_modq(b):
    return int.from_bytes(b, 'little') % q


def point_add(P, Q):
    A, B = (P[1]-P[0]) * (Q[1]-Q[0]) % p, (P[1]+P[0]) * (Q[1]+Q[0]) % p
    C, D = 2 * P[3] * Q[3] * d % p, 2 * P[2] * Q[2] % p
    E, F, G, H = B-A, D-C, D+C, B+A
    return (E*F, G*H, F*G, E*H)


def point_mul(s, P):
    Q = (0, 1, 1, 0)  # Neutral element
    while s > 0:
        if s & 1:
            Q = point_add(Q, P)
        P = point_add(P, P)
        s >>= 1
    return Q


def point_equal(P, Q):
    # x1 / z1 == x2 / z2  <==>  x1 * z2 == x2 * z1
    if (P[0] * Q[2] - Q[0] * P[2]) % p != 0:
        return False
    if (P[1] * Q[2] - Q[1] * P[2]) % p != 0:
        return False
    return True


def recover_x(y, sign):
    if y >= p:
        return None
    x2 = (y*y-1) * modp_inv(d*y*y+1)
    if x2 == 0:
        if sign:
            return None
        else:
            return 0

    # Compute square root of x2
    x = pow(x2, (p+3) // 8, p)
    if (x*x - x2) % p != 0:
        x = x * modp_sqrt_m1 % p
    if (x*x - x2) % p != 0:
        return None

    if (x & 1) != sign:
        x = p - x
    return x


g_y = 4 * modp_inv(5) % p
g_x = recover_x(g_y, 0)
G = (g_x, g_y, 1, g_x * g_y % p)


def point_compress(P):
    zinv = modp_inv(P[2])
    x = P[0] * zinv % p
    y = P[1] * zinv % p
    return int.to_bytes(y | ((x & 1) << 255), 32, "little")


def point_decompress(s):
    if len(s) != 32:
        raise Exception("Invalid input length for decompression")
    y = int.from_bytes(s, "little")
    sign = y >> 255
    y &= (1 << 255) - 1

    x = recover_x(y, sign)
    if x is None:
        return None
    else:
        return (x, y, 1, x*y % p)

## These are functions for manipulating the private key.

def secret_expand(secret):
    if len(secret) != 32:
        raise Exception("Bad size of private key")
    h = sha512(secret)
    a = int.from_bytes(h[:32], "little")
    a &= (1 << 254) - 8
    a |= (1 << 254)
    return (a, h[32:])


def secret_to_public(secret):
    (a, dummy) = secret_expand(secret)
    return point_compress(point_mul(a, G))


def verify(public, msg, signature):
    if len(public) != 32:
        raise Exception("Bad public key length")
    if len(signature) != 64:
        Exception("Bad signature length")
    A = point_decompress(public)
    if not A:
        return False
    Rs = signature[:32]
    R = point_decompress(Rs)
    if not R:
        return False
    s = int.from_bytes(signature[32:], "little")
    if s >= q: return False
    h = sha512_modq(Rs + public + msg)
    sB = point_mul(s, G)
    hA = point_mul(h, A)
    return point_equal(sB, point_add(R, hA))


def _sign(a, prefix, msg):
    A = point_compress(point_mul(a, G))
    r = sha512_modq(prefix + msg)
    R = point_mul(r, G)
    Rs = point_compress(R)
    h = sha512_modq(Rs + A + msg)
    s = (r + h * a) % q
    return Rs + int.to_bytes(s, 32, "little")


def sign(secret, msg):
    a, prefix = secret_expand(secret)
    return _sign(a, prefix, msg)


def get25519Params():
    return (p, q, G)


from stealth25519.ed25519 import get25519Params, _sign, secret_expand, point_compress, sha512, bytes_modq, point_mul, sha256

p, q, G = get25519Params()

class StealthAddressSigner:
    """
    Signs Stealth Addresses for Ed25519-based cryptography.

    Attributes:
        privateSpendKey (PrivateKey): The private spend key.
        privateViewKey (PrivateKey): The private view key.
        bs (int): The expanded private spend key scalar.
        vs (int): The expanded private view key scalar.

    Methods:
        sign(stealthAddress, msg): Sign the given Stealth Address and message.

    Returns:
        bool: True if the transaction is intended for the recipient
    """

    def __init__(self, privateSpendKey, privateViewKey, hash_function = sha512):
        """
        Initialize a StealthAddressSigner instance.

        Args:
            privateSpendKey (PrivateKey): The private spend key.
            privateViewKey (PrivateKey): The private view key.
            hash_function (function): A function that used to generate a hash. This function should get bytes input and return hash bytes. 
        """
        self.hash_function = hash_function

        self.privateSpendKey = privateSpendKey
        self.privateViewKey = privateViewKey
        bs, dummy = secret_expand(privateSpendKey.getPrivateBytes()) # ed25519
        vs, dummy = secret_expand(privateViewKey.getPrivateBytes()) # ed25519
        self.bs = bs
        self.vs = vs

    def sign(self, stealthAddress, msg):
        """Sign the given Stealth Address and message.

        Args:
            stealthAddress (StealthAddress): The Stealth Address to be signed.
            msg (bytes): The message to be signed.

        Returns:
            str: The hexadecimal representation of the signature.
        """

        R, Ps =stealthAddress.R, stealthAddress.Ps
        f = point_compress(point_mul(self.vs, R)) # ed25519
        h = self.hash_function(f)  # ed25519
        h = bytes_modq(h)
        sk = (h + self.bs) % q
        pk = point_mul(sk, G) # ed25519
        pk = point_compress(pk) # ed25519
        if pk!=Ps:
            raise Exception("Not a valid transaction")
    
        prefix = sha256(sk.to_bytes(32, 'little')) # ed25519
        return _sign(sk, prefix, msg) # ed25519




from stealth25519.ed25519 import get25519Params, point_compress, point_decompress, sha512_modq, sha512, bytes_modq, point_mul, point_add
from stealth25519.address import StealthAddress
import random

p, q, G = get25519Params()


class StealthAddressGenerator:
    """
    Generates Stealth Addresses for Ed25519-based cryptography.

    Attributes:
        publicSpendKey (PublicKey): The public spend key.
        publicViewKey (PublicKey): The public view key.
        V (Point): The decompressed point corresponding to the public view key.
        B (Point): The decompressed point corresponding to the public spend key.

    Methods:
        generate(): Generate a Stealth Address.
    """

    def __init__(self, publicSpendKey = None, publicViewKey = None, hash_function = sha512):
        """
        Initialize a StealthAddressGenerator instance.

        Args:
            publicSpendKey (PublicKey): The public spend key.
            publicViewKey (PublicKey): The public view key.
            hash_function (function): A function that used to generate a hash. This function should get bytes input and return hash bytes. 
        """
        self.hash_function = hash_function

        self.publicSpendKey = publicSpendKey
        self.publicViewKey = publicViewKey
        self.V = point_decompress(publicViewKey.getPublicBytes()) # ed25519
        self.B = point_decompress(publicSpendKey.getPublicBytes()) # ed25519
        
    def generate(self):
        """Generate a Stealth Address."""

        r = sha512_modq(random.randbytes(32))
        R = point_mul(r, G)
        Rs = point_compress(R)
    
        f = point_compress(point_mul(r, self.V))
        h = self.hash_function(f) # ed25519
        h = bytes_modq(h)
        P = point_add(point_mul(h, G), self.B)
        Ps = point_compress(P)
        
        return StealthAddress(Rs, Ps)





from stealth25519.ed25519 import get25519Params, secret_expand, point_compress, sha512, bytes_modq, point_add, point_mul

p, q, G = get25519Params()

class StealthAddressVerifier:
    """
    Verifies Stealth Addresses for Ed25519-based cryptography.

    Attributes:
        privateViewKey (PrivateKey): The private view key.
        publicSpendKey (PublicKey): The public spend key.
        vs (int): The expanded private view key scalar.
        B (Point): The decompressed point corresponding to the public spend key.

    Methods:
        verify(stealthAddress): Verify the given Stealth Address.
    """

    def __init__(self, privateViewKey, publicSpendKey, hash_function = sha512):
        """
        Initialize a StealthAddressVerifier instance.

        Args:
            privateViewKey (PrivateKey): The private view key.
            publicSpendKey (PublicKey): The public spend key.
            hash_function (function): A function that used to generate a hash. This function should get bytes input and return hash bytes. 
        """
        self.hash_function = hash_function

        self.privateViewKey = privateViewKey
        self.publicSpendKey = publicSpendKey
        vs, dummy = secret_expand(privateViewKey.getPrivateBytes()) # ed25519
        self.vs = vs
        self.B = publicSpendKey.getPointOnCurve()


    def verify(self, stealthAddress):
        """Verify the given Stealth Address.

        Args:
            stealthAddress (StealthAddress): The Stealth Address to be verified.
        """

        Rs, Ps = stealthAddress.Rs, stealthAddress.Ps
        f = point_compress(point_mul(self.vs, stealthAddress.R)) # ed25519
        h_ = self.hash_function(f) # ed25519
        h_ = bytes_modq(h_)
        P_ = point_add(point_mul(h_, G), self.B) # ed25519
        Ps_ = point_compress(P_) # ed25519
        return Ps==Ps_




from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='stealth25519',
    version='1.0.9',
    description='The Stealth25519 Python package provides functionality for the generation, verification, and signing of stealth addresses using the standard ed25519 algorithm with the ECDH protocol.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=[],
    tests_require=[],
    package_data={'': ['LICENSE']},
    include_package_data=True,
    license='MIT',
    project_urls={
        'Source': 'https://github.com/khqnn/stealth-py',
    },
    entry_points={
        'console_scripts': [],
    },
)
